--- drweb-configure samba module
-- @copyright 2018 Dr.Web Ltd.

--- Separator to be inserted before and after the new field
comm_separator = "========================"

--- Comment to be inserted before the new field
comm = "Added by drweb-configure"

--- Samba module name
module = "smb_spider"

--- Initialize augeas object, load samba config and set `smbconf' augeas variable
-- containing path to it in augeas tree.
-- @param config path to smb.conf
-- @return augeas object
local function init(config)
  local augeas = require("augeas")
  local aug = augeas.init({ noautoload = true, save_backup = true, no_load = true, no_modl_autoload = true })

  aug:set("/augeas/load/Samba/lens", "Samba.lns")
  aug:set("/augeas/load/Samba/incl", config)
  aug:load()

  if aug:error() > 0 then
    io.stderr:write(aug:error_message(), "\n")
    aug:close()
    return nil
  end

  aug:defvar("smbconf", "/files" .. config)
  return aug
end

--- Create new node and additional comment lines around it if needed.
-- Assume the node with the specified label does not exist.
-- @param aug augeas object
-- @param target augeas path of parent node
-- @param name new node label
-- @return augeas path of new node
local function new_node(aug, target, name)
  local startblock = aug:match(target .. "/#comment[. = '" .. comm .. "']")[1]
  if not startblock then
    -- insert after last non-comment line
    local last = aug:match(target .. "/*[label() != '#comment'][last()]")[1]
    assert(last, target .. " is empty!")
    local ncomms = aug:matches(target .. "/#comment[following-sibling::*[label() != '#comment']]")
    local last_label = aug:label(last)
    aug:insert(last, "#comment", false)
    aug:insert(last, name, false)
    aug:insert(last, "#comment", false)
    aug:insert(last, "#comment", false)
    aug:set(target .. "/#comment[" .. (ncomms + 1) .. "]", comm_separator)
    aug:set(target .. "/#comment[" .. (ncomms + 2) .. "]", comm)
    aug:set(target .. "/#comment[" .. (ncomms + 3) .. "]", comm_separator)
  else
    aug:insert(startblock, name, false)
  end
  return target .. "/" .. name
end

--- Remove the comment block added by `new_node` if it is empty.
-- @arg aug augeas object
-- @arg target augeas path to the target
local function target_cleanup(aug, target)
  local commpos = aug:matches(target .. "/*[following-sibling::#comment[. = '" .. comm .. "']]")
  if commpos > 0 then
    if
      aug:label(target .. "/*[" .. commpos .. "]") == "#comment" and
      aug:get(target .. "/*[" .. commpos .. "]") == comm_separator and
      aug:label(target .. "/*[" .. commpos + 2 .. "]") == "#comment" and
      aug:get(target .. "/*[" .. commpos + 2 .. "]") == comm_separator
    then
      aug:rm(target .. "/*[" .. commpos + 2 .. "]")
      aug:rm(target .. "/*[" .. commpos + 1 .. "]")
      aug:rm(target .. "/*[" .. commpos .. "]")
    end
  end
end

--- Find a node in augeas tree, containing `vfs objects' or `vfs object' parameter
-- of the given samba share.
-- @param aug augeas object
-- @param target augeas path to samba share
-- @param create whether the node will be created if does not exist
-- @return augeas path to `vfs objects'
local function vfs_objects_path(aug, target, create)
  for i, path in next, {target .. "/vfs objects", target .. "/vfs object"}, nil do
    if aug:matches(path) > 0 then
      return path
    end
  end
  if create then
    return new_node(aug, target, "vfs objects")
  end
  return nil
end

--- Check if the share is protected.
-- @param aug augeas object
-- @param path augeas path to the share
local function is_protected(aug, path)
  local node = vfs_objects_path(aug, path, false)
  if node ~= nil then
    local val = aug:get(node)
    if val and string.len(val) then
      if string.match(val, "%f[^\0\t ]" .. module .. "%f[\0\t ]") then
        return true
      end
    end
  end
  return false
end

--- Print configured shares in samba config in form `+|- [name]: comment (path)'.
-- @param aug augeas object
local function list(aug)
  local targets = aug:match("$smbconf/target")
  for i, node in ipairs(targets) do
    local name = aug:get(node)
    if name ~= "global" then
      local comm = aug:get(node .. "/comment")
      local path = aug:get(node .. "/path")
      if is_protected(aug, node) then
        io.write("+ ")
      else
        io.write("- ")
      end
      io.write("[" .. name .. "]")
      if comm then
        io.write(": " .. comm)
      end
      if path then
        io.write(" (" .. path .. ")")
      end
      io.write("\n")
    end
  end
end

--- Get all configured share paths.
-- @param aug augeas object
-- @return table of augeas paths indexed by share names
local function all_targets(aug)
  local targets = aug:match("$smbconf/target")
  local result = {}
  for i, node in ipairs(targets) do
    local name = aug:get(node)
    if name ~= "global" then
      result[name] = node
    end
  end
  return result
end

--- Find paths in augeas tree corresponding to samba share names.
-- Name `/all' is treated as special and means all configured shares.
-- @param aug augeas object
-- @param targets table of share names
-- @return table of augeas paths indexed by share names
local function resolve_targets(aug, targets)
  local result = {}

  for i, name in ipairs(targets) do
    if name == "/all" then
      return all_targets(aug)
    end
    local t = aug:match("$smbconf/target[. = \"" .. name .. "\"]")
    if #t == 0 then
      io.stderr:write(string.format("Error: target [%s] not found\n", name))
    else
      result[name] = t[1]
    end
  end

  return result
end

--- Get names of vfs objects in global section (except "smb_spider")
-- @param aug augeas object
-- @return table of strings
local function get_global_vfs_names(aug)
  local targets = aug:match("$smbconf/target")
  local result = {}
  for i, node in ipairs(targets) do
    local name = aug:get(node)
    if name == "global" then
      local vfs_node = vfs_objects_path(aug, node, false)
      if vfs_node then
        local val = aug:get(vfs_node)
        if val and string.len(val) then
          for m in string.gmatch(val, "%f[^\0\t ][%w_-]+%f[\0\t ]") do
            if m ~= module then
              table.insert(result, m)
            end
          end
        end
      end
      break
    end
  end
  return result
end

--- Enable protection of given samba shares.
-- @param aug augeas object
-- @param targets table of samba share names
local function protect(aug, targets, global_names)
  local add_names = global_names
  table.insert(add_names, module)
  local paths = resolve_targets(aug, targets)
  for name, target in pairs(paths) do
    local node = vfs_objects_path(aug, target, true)
    local val = aug:get(node)
    if val and string.len(val) then
      for _, m in pairs(add_names) do
        if not string.match(val, "%f[^\0\t ]" .. m .. "%f[\0\t ]") then
          val = val .. " " .. m
        end
      end
    else
      val = table.concat(add_names, " ")
    end
    aug:set(node, val)
  end
end

--- Disable protection of given samba shares.
-- @param aug augeas object
-- @param targets table of samba share names
local function unprotect(aug, targets, global_names)
  local paths = resolve_targets(aug, targets)
  for name, target in pairs(paths) do
    local node = vfs_objects_path(aug, target, false)
    if node then
      local val = aug:get(node)
      if val and string.len(val) then
        local newval = {};
        for m in string.gmatch(val, "%f[^\0\t ][%w_-]+%f[\0\t ]") do
          if m ~= module then
            local found = false
            for _, r in pairs(global_names) do
              if m == r then
                found = true
                break
              end
            end
            if not found then
              table.insert(newval, m)
            end
          end
        end
        if #newval > 0 then
          aug:set(node, table.concat(newval, " "))
        else
          aug:rm(node)
          target_cleanup(aug, target)
        end
      end
    end
  end
end

--- Print usage information
local function usage()
  io.write(string.format("Usage: %s %s [{+|-}<share> ... | {+|-}/all] [<conffile>] [add_symlink] [remove_symlink]\n", arg[-1], arg[0]))
end

--- Find samba bindir
-- @return file path
local function guess_bindir()
  local unistd = require("posix.unistd")
  local smbd = {}
  for i, prefix in ipairs({"/usr", "/usr/local", "/usr/local/samba", "/opt/samba"}) do
    for j, bindir in ipairs({"bin", "sbin"}) do
      local b = prefix .. "/" .. bindir .. "/smbd"
      if unistd.access(b, "x") then
        table.insert(smbd, b)
      end
    end
  end
  return smbd
end

-- Check if Samba is installed
local function check_inst_smbd()
  local lstat = require "posix.sys.stat".lstat
  local smbd = guess_bindir()
  for i, binary in ipairs(smbd) do
    local exist_binary = lstat(binary)
    if exist_binary == nil then
      return nil
    else
      return true
    end
  end
end

--- Find samba configuration file
local function guess_conffile()
  local unistd = require("posix.unistd")
  local smbd = guess_bindir()
  local conffiles = {}
  for i, binary in ipairs(smbd) do
    local out = io.popen("env -u LD_LIBRARY_PATH " .. binary .. " -b")
    for line in out:lines() do
      local c = string.match(line, "[ \t]*CONFIGFILE:[ \t]+(.+)")
      if c and unistd.access(c, "r") then
        table.insert(conffiles, c)
        break
      end
    end
    out:close()
  end
  if #conffiles == 1 then
    return conffiles[1]
  elseif #conffiles == 0 then
    io.stderr:write("Error: Cannot find the Samba configuration file.\n")
    io.stderr:write("       Please specify path in the command line.\n")
  else
    io.stderr:write("Error: Cannot guess a Samba configuration file.\n")
    io.stderr:write("       Candidates:\n")
    for i, c in ipairs(conffiles) do
      io.stderr:write(string.format("         %s\n", c))
    end
    io.stderr:write("       Please specify one of them in the command line.\n")
  end
  return nil
end

-- Get drweb libdir for samba
local function get_libdir()
  local glob = require 'posix.glob'.glob
  local dir = require 'posix.dirent'.dir
  local dir_find = pcall(dir, '/opt/drweb.com/lib')
  local libdir
  if dir_find == true then
    for i, libdir_path in ipairs(glob("/opt/drweb.com/lib*/*/samba", 0)) do
      libdir = libdir_path
    end
  else
    for i, libdir_path in ipairs(glob("/usr/local/libexec/drweb.com/lib*/samba", 0)) do
      libdir = libdir_path
    end
  end
  return libdir
end

-- Get all samba_libdir
-- Get all Samba version
local function get_ver_smb_libdir()
  local smbd = guess_bindir()
  local smbd_libdir = {}
  local smbd_libdir1= {}
  local ver_samba = {}
  local slib, smbvers
  for i, binary in ipairs(smbd) do
    local out, out1 = io.popen("env -u LD_LIBRARY_PATH " .. binary .. " -b"), io.popen("env -u LD_LIBRARY_PATH " .. binary .. " -V")
    for line in out:lines() do
      local slib = string.match(line,"%s*MODULESDIR:%s+(.+)$")
      if slib then
        table.insert(smbd_libdir, slib)
      end
      local slib1 = string.match(line,"%s*LIBDIR:%s+(.+)$")
      if slib1 then
        table.insert(smbd_libdir1, slib1)
      end
    end
    for line1 in out1:lines() do
      local smbvers = line1:gsub("^Version (%d+%.%d+)%..*", "%1.0")
      if smbvers then
        table.insert(ver_samba, smbvers)
      end
    end
    out:close()
  end
  if smbd_libdir == nil and smbd_libdir1 then
    smbd_libdir = smbd_libdir1
  end
  return ver_samba, smbd_libdir
end

-- create a symlink and check if the file or symlink exists
local function add_symlink()
  local unistd = require("posix.unistd")
  local lstat = require "posix.sys.stat".lstat
  local mkdir = require "posix".mkdir
  local eexist = require "posix.errno"['EEXIST']
  local readlink = require 'posix.unistd'.readlink
  local num1, smbd_libdir = get_ver_smb_libdir()
  local libdir = get_libdir() .. "/libsmb_spider.so."
  local number_tab = select(1,num1)
  local smbd_libdir_tab = select(2,get_ver_smb_libdir())
  assert(#num1 == #smbd_libdir)
  if number_tab[1] and number_tab[2] then
    local libdir_path1, libdir_path2 = lstat(libdir .. number_tab[1]), lstat(libdir .. number_tab[2])
    if libdir_path1 and libdir_path2 then
      for i, number in ipairs(num1) do
        smb_libdir = smbd_libdir[i]
        local libdir_p1 = libdir .. number
        local smbd_libdir_vfs = smb_libdir .. "/vfs"
        local smbd_libdir_link = smbd_libdir_vfs .. "/smb_spider.so"
        local symlink = readlink(smbd_libdir_link)
        local reg_file = lstat(smbd_libdir_link)
        local res, msg, err = mkdir(smbd_libdir_vfs)
        if not res and err ~= eexist then
          print("Error creating vfs directory: " .. msg .. "\n")
        end
        if symlink then
          print("Found old symlink in " .. smbd_libdir_link .. ". Removing...\n")
          unistd.unlink(smbd_libdir_link)
        elseif not symlink then
          if reg_file then
            io.stderr:write(string.format(smbd_libdir_link .. " is a regular file. Cannot create symlink.\n"))
            return false
          end
        end
        if unistd.link(libdir_p1, smbd_libdir_link, true) then
          print("Creating symlink " .. smbd_libdir_link .. "-->" .. libdir_p1)
          print("Please update your config.")
        else
          print("Сannot create link " .. smbd_libdir_link .. " (Permission denied)")
        end
      end
      return true
    elseif libdir_path1 == nil and libdir_path2 then
      print("Could not determine SAMBA version" .. "-" .. number_tab[1])
      libdir_p1 = libdir .. number_tab[2]
      smbd_libdir_vfs = smbd_libdir_tab[2] .. "/vfs"
    elseif libdir_path2 == nil and libdir_path1 then
      print("Could not determine SAMBA version" .. "-" .. number_tab[2])
      libdir_p1 = libdir .. number_tab[1]
      smbd_libdir_vfs = smbd_libdir_tab[1] .. "/vfs"
    elseif libdir_path1 == nil and libdir_path2 == nil then
      print("Could not determine SAMBA version" .. "-" .. number_tab[1] .. "  " .. number_tab[2])
      return false
    end
  elseif number_tab[1] then
    local libdir_path1 = lstat(libdir .. number_tab[1])
    number_tab[2] = ''
    if libdir_path1 then
      libdir_p1 = libdir .. number_tab[1]
      smbd_libdir_vfs = smbd_libdir_tab[1] .. "/vfs"
    elseif libdir_path1 == nil then
      print("Could not determine SAMBA version" .. "-" .. number_tab[1])
      return false
    end
  elseif number_tab[2] then
    local libdir_path2 = lstat(libdir .. number_tab[2])
    number_tab[1] = ''
    if libdir_path2 then
      libdir_p1 = libdir .. number_tab[2]
      smbd_libdir_vfs = smbd_libdir_tab[2] .. "/vfs"
    elseif libdir_path2 == nil then
      print("Could not determine SAMBA version" .. "-" .. number_tab[2])
      return false
    end
  end
  local smbd_libdir_link = smbd_libdir_vfs .. "/smb_spider.so"
  local symlink = readlink(smbd_libdir_link)
  local reg_file = lstat(smbd_libdir_link)
  local res, msg, err = mkdir(smbd_libdir_vfs)
  if not res and err ~= eexist then
    print("Error creating vfs directory: " .. msg .. "\n")
  end
  if symlink then
    print("Found old symlink in " .. smbd_libdir_link .. ". Removing...\n")
    unistd.unlink(smbd_libdir_link)
  elseif not symlink then
    if reg_file then
      io.stderr:write(string.format(smbd_libdir_link .. " is a regular file. Cannot create symlink.\n"))
      return false
    end
  end
  if unistd.link(libdir_p1, smbd_libdir_link, true) then
    print("Creating symlink " .. smbd_libdir_link .. "-->" .. libdir_p1)
    print("Please update your config.")
  else
    print("Сannot create link " .. smbd_libdir_link .. " (Permission denied)")
  end
end

local function remove_symlink()
  local unistd = require("posix.unistd")
  local readlink = require 'posix.unistd'.readlink
  local lstat = require "posix.sys.stat".lstat
  local smbd_libdir = select(2,get_ver_smb_libdir())
  for i, smbd_libdir_p in ipairs(smbd_libdir) do
    smbd_libdir = smbd_libdir_p .. "/vfs" .. "/smb_spider.so"
    local symlink = readlink(smbd_libdir)
    local reg_file = lstat(smbd_libdir)
    if symlink then
      print("Removing link " .. smbd_libdir)
      unistd.unlink(smbd_libdir)
      local symlink1 = readlink(smbd_libdir)
      if symlink1 then
        print("Сannot remove link " .. smbd_libdir .. " (Permission denied)")
      else
        print("Please update your config.")
      end
    elseif not  symlink then
      if reg_file then
        print("Removing link " .. smbd_libdir)
        io.stderr:write(string.format(smbd_libdir .. " is a regular file. Cannot remove symlink.\n"))
        print("Please update your config.")
      end
    end
  end
end

function samba_main()
  local conffile = nil
  local p_shares = {}
  local u_shares = {}
  local status = 0
  local exist_binary = check_inst_smbd()

  if exist_binary == nil then
    io.stderr:write("Error: Cannot find any smbd.\n")
    os.exit(0)
  else
    for i, a in ipairs(arg) do
      if string.sub(a, 1, 1) == "+" then
        table.insert(p_shares, string.sub(a, 2))
      elseif string.sub(a, 1, 1) == "-" then
        table.insert(u_shares, string.sub(a, 2))
      elseif a == "add_symlink" then
        add_symlink()
        return 0
      elseif a == "remove_symlink" then
        remove_symlink()
        return 0
      elseif conffile == nil then
        local stdlib = require("posix.stdlib")
        conffile = stdlib.realpath(a)
        if not conffile then
          io.stderr:write(string.format("Error: No such file - `%s'\n", a))
          usage()
          return 1
        end
      else
        usage()
        return 1
      end
    end

    if conffile == nil then
      conffile = guess_conffile()
    end
    if conffile == nil then
      return 2
    end

    local aug = init(conffile)
    if not aug then
      io.stderr:write("Error: Cannot initialize augeas\n")
      return 3
    end

    if #p_shares > 0 then
      local cf = io.popen("drweb-ctl cfshow --value smbspider.start")
      local start = cf:read()
      cf:close()
      if not (start == "Yes" or os.execute("drweb-ctl cfset smbspider.start yes")) then
        io.stderr:write("Error: Cannot start Dr.Web SpIDer Guard for SMB.\n")
        io.stderr:write("       Not enabling protection to prevent Samba failure.\n\n")
        p_shares = {}
        status = 4
      end
    end

    local global_names = get_global_vfs_names(aug)
    unprotect(aug, u_shares, global_names)
    protect(aug, p_shares, global_names)

    if aug:save() ~= 0 then
      io.stderr:write(string.format("Error: Cannot save changes to `%s'.\n", conffile))
      aug:close()
      return 5
    end

    if #p_shares + #u_shares > 0 then
      if os.execute("env -u LD_LIBRARY_PATH smbcontrol smbd reload-config") then
        io.write("Configuration changes applied\n\n")
      else
        io.write("Please reload Samba (smbd) for changes to take effect\n\n")
      end
    end

    list(aug)

    aug:close()

  end
  return status
end

function samba_help()
  usage()
  io.stderr:write("\n")
  io.stderr:write("    +<share>   Enable protection for specified Samba share.\n")
  io.stderr:write("    -<share>   Disable protection for specified Samba share.\n")
  io.stderr:write("    +/all      Enable protection for all Samba shares.\n")
  io.stderr:write("    -/all      Disable protection for all Samba shares.\n")
  io.stderr:write("    <conffile> Path to the Samba configuration file (smb.conf). If not specified, program will try to guess it automatically.\n")
  io.stderr:write("    add_symlink   Create symlink for Samba version.\n")
  io.stderr:write("    remove_symlink  Remove symlink.\n")
  io.stderr:write("\n")
  io.stderr:write("Output legend:\n")
  io.stderr:write("STATUS [NAME]: COMMENT (PATH)\n")
  io.stderr:write("    STATUS  `+' if protection is enabled, `-' if protection is disabled.\n")
  io.stderr:write("    NAME    Share name as it is specified in `smb.conf'.\n")
  io.stderr:write("    COMMENT Share comment as it is specified in `smb.conf' (may be empty).\n")
  io.stderr:write("    PATH    Path to the shared directory as it is specified in `smb.conf' (may be empty).\n")
end
